from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from fluent.syntax import FluentParser, ast

from ftl_extract.matcher import FluentKey

if TYPE_CHECKING:
    from fluent.syntax.ast import Resource

    from ftl_extract.utils import ExtractionStatistics


def import_from_ftl(
    *,
    path: Path,
    locale: str,
    parser: FluentParser,
) -> tuple[dict[str, FluentKey], dict[str, FluentKey], Resource, list[FluentKey]]:
    """Import `FluentKey`s from FTL."""
    ftl_keys: dict[str, FluentKey] = {}
    terms: dict[str, FluentKey] = {}
    leave_as_is = []

    resource = parser.parse(path.read_text(encoding="utf-8"))

    for position, entry in enumerate(resource.body, start=0):
        if isinstance(entry, ast.Message):
            ftl_keys[entry.id.name] = FluentKey(
                code_path=Path(),
                key=entry.id.name,
                translation=entry,
                path=path,
                locale=locale,
                position=position,
            )
        elif isinstance(entry, ast.Term):
            terms[entry.id.name] = FluentKey(
                code_path=Path(),
                key=entry.id.name,
                translation=entry,
                path=path,
                locale=locale,
                position=position,
            )
        else:
            leave_as_is.append(
                FluentKey(
                    code_path=Path(),
                    key="",
                    translation=entry,
                    path=path,
                    locale=locale,
                    position=position,
                ),
            )

    return ftl_keys, terms, resource, leave_as_is


def import_ftl_from_dir(
    *,
    path: Path,
    locale: str,
    statistics: ExtractionStatistics | None = None,
) -> tuple[dict[str, FluentKey], dict[str, FluentKey], list[FluentKey]]:
    """Import `FluentKey`s from directory of FTL files."""
    ftl_files = (path / locale).rglob("*.ftl") if path.is_dir() else [path]
    stored_ftl_keys: dict[str, FluentKey] = {}
    stored_terms: dict[str, FluentKey] = {}
    stored_leave_as_is_keys = []
    parser = FluentParser(with_spans=True)

    for ftl_file in ftl_files:
        keys, terms, _, leave_as_is_keys = import_from_ftl(
            path=ftl_file,
            locale=locale,
            parser=parser,
        )
        stored_ftl_keys.update(keys)
        stored_terms.update(terms)
        stored_leave_as_is_keys.extend(leave_as_is_keys)

        if statistics:
            statistics.ftl_files_count[locale] += 1

    return stored_ftl_keys, stored_terms, stored_leave_as_is_keys



# Stub
from typing import Any


class I18nContext:
    def get(self, *_, **__) -> None: ...

    def __getattr__(self, item: str) -> Any: ...

    def __call__(self, *_, **__) -> None: ...


class Mock:
    cls_i18n: I18nContext

    def __init__(self, i18n: I18nContext) -> None:
        self.i18n = i18n

    def self_i18n(self) -> None:
        self.i18n.self.key(some_kwarg="...", _path="classlike.ftl")
        self.i18n.get("self-get-key", some_kwarg="...", _path="classlike.ftl")

    @classmethod
    def cls_i18n(cls) -> None:
        cls.cls_i18n.cls.key(some_kwarg="...", _path="classlike.ftl")
        cls.cls_i18n.get("cls-get-key", some_kwarg="...", _path="classlike.ftl")



from __future__ import annotations

import ast
from dataclasses import dataclass, field
from math import inf
from pathlib import Path
from typing import TYPE_CHECKING, cast

from fluent.syntax import ast as fluent_ast

from ftl_extract.const import (
    DEFAULT_I18N_KEYS,
    DEFAULT_IGNORE_ATTRIBUTES,
    DEFAULT_IGNORE_KWARGS,
    GET_LITERAL,
    PATH_LITERAL,
)
from ftl_extract.exceptions import (
    FTLExtractorDifferentPathsError,
    FTLExtractorDifferentTranslationError,
)
from ftl_extract.utils import to_json_no_span

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass
class FluentKey:
    """
    Dataclass for storing information about key and its translation.

    :param code_path: Path to .py file where key was found.
    :type code_path: Path
    :param key: Key that will be used to get translation.
    :type key: str | None
    :param translation: Translation of key.
    :type translation: str | None
    :param path: Path to .ftl file where key will be stored.
    :type path: Path
    :param locale: Locale of translation. When extracting from .py file, it will not be needed.
    :type locale: str | None
    """

    code_path: Path
    key: str
    translation: fluent_ast.EntryType
    path: Path
    locale: str | None = field(default=None)
    position: int | float = field(default=inf)
    depends_on_keys: set[str] = field(default_factory=set)

    def __repr__(self) -> str:
        return (
            f"FluentKey("
            f"code_path={self.code_path},"
            f"key={self.key},"
            f"path={self.path},"
            f"locale={self.locale},"
            f"position={self.position},"
            f"translation={self.translation.to_json(fn=to_json_no_span)}"
            f")"
        )

    def __str__(self) -> str:
        return (
            f"FluentKey(\n"
            f"\tcode_path={self.code_path},\n"
            f"\tkey={self.key},\n"
            f"\tpath={self.path},\n"
            f"\tlocale={self.locale},\n"
            f"\tposition={self.position},\n"
            f"\ttranslation={self.translation.to_json(fn=to_json_no_span)}\n"
            f")"
        )


class I18nMatcher(ast.NodeVisitor):
    def __init__(
        self,
        code_path: Path,
        default_ftl_file: Path,
        i18n_keys: Iterable[str] = DEFAULT_I18N_KEYS,
        i18n_keys_prefix: Iterable[str] = (),
        ignore_attributes: Iterable[str] = DEFAULT_IGNORE_ATTRIBUTES,
        ignore_kwargs: Iterable[str] = DEFAULT_IGNORE_KWARGS,
    ) -> None:
        """

        :param code_path: Path to .py file where visitor will be used.
        :type code_path: Path
        :param default_ftl_file: Default name of FTL file.
        :type default_ftl_file: Path
        :param i18n_keys: Name of function that is used to get translation. Default is ("i18n",).
        :type i18n_keys: Iterable[str]
        :param i18n_keys_prefix: Prefix names of function that is used to get translation.
        :type i18n_keys_prefix: Iterable[str]
        :param ignore_attributes: Ignore attributes, like `i18n.set_locale`.
        :type ignore_attributes: Iterable[str]
        :param ignore_kwargs: Ignore kwargs, like `when` from
        `aiogram_dialog.I18nFormat(..., when=...)`.
        :type ignore_kwargs: Iterable[str]
        """
        self.code_path = code_path
        self.i18n_keys = frozenset(i18n_keys)
        self.i18n_keys_prefix = frozenset(i18n_keys_prefix)
        self.ignore_attributes = frozenset(ignore_attributes)
        self.ignore_kwargs = frozenset(ignore_kwargs)
        self.default_ftl_file = default_ftl_file
        self.fluent_keys: dict[str, FluentKey] = {}

    def visit_Call(self, node: ast.Call) -> None:
        if isinstance(node.func, ast.Attribute):
            self._process_attribute_call(node)

        elif isinstance(node.func, ast.Name) and node.func.id in self.i18n_keys:
            self._process_name_call(node)

        self.generic_visit(node)

    def _process_attribute_call(self, node: ast.Call) -> None:
        attr: ast.Attribute | ast.expr = node.func
        attrs = []
        while isinstance(attr, ast.Attribute):
            attrs.append(attr.attr)
            attr = attr.value

        if isinstance(attr, ast.Name):
            self._process_attribute_name_call(node, attr, attrs)

    def _process_name_call(self, node: ast.Call) -> None:
        if not node.args or not isinstance(node.args[0], ast.Constant):
            return

        fluent_key = self._create_fluent_key(
            node=node, key=cast(str, cast(ast.Constant, node.args[0]).value)
        )
        self._add_fluent_key(new_fluent_key=fluent_key)

    def _process_attribute_name_call(
        self,
        node: ast.Call,
        attr: ast.Name,
        attrs: list[str],
    ) -> None:
        if attr.id in self.i18n_keys:
            self._process_i18n_key_call(node, attrs)

        elif attr.id in self.i18n_keys_prefix and attrs and attrs[-1] in self.i18n_keys:
            # Remove the last attribute to handle cases where the prefix key is followed by a
            # valid i18n key.
            del attrs[-1]
            self._process_i18n_key_call(node, attrs)

    def _process_i18n_key_call(self, node: ast.Call, attrs: list[str]) -> None:
        if len(attrs) == 1 and attrs[0] == GET_LITERAL:
            self._process_i18n_key_call_get_literal(node, attrs)

        else:
            self._process_i18n_key_call_attrs(node, attrs)

    def _process_i18n_key_call_get_literal(self, node: ast.Call, attrs: list[str]) -> None:
        if not node.args:
            return  # Skip if no args

        # Add the first arg as the translation key
        attrs.clear()
        if isinstance(arg := node.args[0], ast.Constant):
            key = cast(str, cast(ast.Constant, arg).value)

            fluent_key = self._create_fluent_key(node=node, key=key)
            self._add_fluent_key(new_fluent_key=fluent_key)

    def _process_i18n_key_call_attrs(self, node: ast.Call, attrs: list[str]) -> None:
        if attrs[-1] in self.ignore_attributes:
            return

        fluent_key = self._create_fluent_key(node=node, key="-".join(reversed(attrs)))
        self._add_fluent_key(new_fluent_key=fluent_key)

    def _create_fluent_key(self, *, node: ast.Call, key: str) -> FluentKey:
        fluent_key = FluentKey(
            code_path=self.code_path,
            key=key,
            translation=fluent_ast.Message(
                id=fluent_ast.Identifier(name=key),
                value=fluent_ast.Pattern(elements=[fluent_ast.TextElement(value=key)]),
            ),
            path=self.default_ftl_file,
        )

        keywords = sorted(node.keywords, key=lambda keyword: keyword.arg or "")

        for kw in keywords:
            if kw.arg == PATH_LITERAL:
                if kw.value is not None and isinstance(kw.value, ast.Constant):
                    fluent_key.path = Path(cast(str, cast(ast.Constant, kw.value).value))
            elif isinstance(kw.arg, str):
                if kw.arg in self.ignore_kwargs:
                    continue

                cast(
                    fluent_ast.Pattern,
                    cast(fluent_ast.Message, fluent_key.translation).value,
                ).elements.append(
                    fluent_ast.Placeable(
                        expression=fluent_ast.VariableReference(
                            id=fluent_ast.Identifier(name=kw.arg),
                        ),
                    ),
                )

        return fluent_key

    def _add_fluent_key(self, *, new_fluent_key: FluentKey) -> None:
        if new_fluent_key.key in self.fluent_keys:
            if self.fluent_keys[new_fluent_key.key].path != new_fluent_key.path:
                raise FTLExtractorDifferentPathsError(
                    new_fluent_key.key,
                    new_fluent_key.path,
                    self.fluent_keys[new_fluent_key.key].path,
                )
            if not self.fluent_keys[new_fluent_key.key].translation.equals(
                new_fluent_key.translation,
            ):
                raise FTLExtractorDifferentTranslationError(
                    new_fluent_key.key,
                    cast(fluent_ast.Message, new_fluent_key.translation),
                    cast(fluent_ast.Message, self.fluent_keys[new_fluent_key.key].translation),
                )

        else:
            self.fluent_keys[new_fluent_key.key] = new_fluent_key



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypedDict, cast

from fluent.syntax.serializer import serialize_pattern

if TYPE_CHECKING:
    from ftl_extract.stub.visitor import Message

METADATA_DICT_KEY: Literal["$_meta$"] = "$_meta$"


class Metadata(TypedDict, total=False):
    args: list[str]
    translation: str


def generate_tree(fluent_messages: dict[str, Message]) -> dict[str, dict[str, Any]]:
    tree: dict[str, dict[str, Any]] = {}

    for key, message in fluent_messages.items():
        key_parts = key.split("-")
        key_parts_len = len(key_parts)

        inner_tree = tree
        for index, key_part in enumerate(key_parts, start=1):
            if index == key_parts_len:
                if key_part in inner_tree:
                    cast(dict[str, Any], inner_tree[key_part])[METADATA_DICT_KEY] = Metadata(
                        args=message.kwargs,
                        translation=serialize_pattern(message.fluent_message.value).strip(),
                    )
                else:
                    inner_tree.setdefault(
                        key_part,
                        {
                            METADATA_DICT_KEY: Metadata(
                                args=message.kwargs,
                                translation=serialize_pattern(message.fluent_message.value).strip(),
                            ),
                        },
                    )
            else:
                inner_tree = inner_tree.setdefault(key_part, {})

    return tree



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from fluent.syntax import ast

from ftl_extract.exceptions import FTLExtractorCantFindReferenceError, FTLExtractorCantFindTermError

if TYPE_CHECKING:
    from ftl_extract.matcher import FluentKey


def _extract_kwargs_from_variable_reference(
    *,
    variable_reference: ast.VariableReference,
    kwargs: set[str],
) -> None:
    kwargs.add(variable_reference.id.name)


def _extract_kwargs_from_selector_expression(
    *,
    key: FluentKey,
    selector_expression: ast.SelectExpression,
    kwargs: set[str],
    terms: dict[str, FluentKey],
    all_fluent_keys: dict[str, FluentKey],
    depend_keys: set[str] | None = None,
) -> None:
    if isinstance(selector_expression.selector, ast.VariableReference):
        _extract_kwargs_from_variable_reference(
            variable_reference=selector_expression.selector,
            kwargs=kwargs,
        )

    for variant in selector_expression.variants:
        for element in variant.value.elements:
            if isinstance(element, ast.Placeable):
                _extract_kwargs_from_placeable(
                    key=key,
                    placeable=element,
                    kwargs=kwargs,
                    terms=terms,
                    all_fluent_keys=all_fluent_keys,
                    depend_keys=depend_keys,
                )


def _extract_kwargs_from_message_reference(
    *,
    key: FluentKey,
    message_reference: ast.MessageReference,
    kwargs: set[str],
    terms: dict[str, FluentKey],
    all_fluent_keys: dict[str, FluentKey],
    depend_keys: set[str] | None = None,
) -> None:
    reference_key = all_fluent_keys.get(message_reference.id.name, None)

    if not reference_key:
        raise FTLExtractorCantFindReferenceError(
            key=key.key,
            key_path=key.path,
            reference_key=message_reference.id.name,
        )

    kwargs.update(
        extract_kwargs(
            key=reference_key,
            terms=terms,
            all_fluent_keys=all_fluent_keys,
            depend_keys=depend_keys,
        ),
    )


def _extract_kwargs_from_term_reference(
    *,
    key: FluentKey,
    term_expression: ast.TermReference,
    kwargs: set[str],
    terms: dict[str, FluentKey],
    all_fluent_keys: dict[str, FluentKey],
) -> None:
    term = terms.get(term_expression.id.name, None)

    if not term:
        raise FTLExtractorCantFindTermError(
            key=key.key,
            locale=cast(str, key.locale),
            key_path=key.path,
            term_key=term_expression.id.name,
        )

    kwargs.update(extract_kwargs(key=term, terms=terms, all_fluent_keys=all_fluent_keys))


def _extract_kwargs_from_placeable(
    *,
    key: FluentKey,
    placeable: ast.Placeable,
    kwargs: set[str],
    terms: dict[str, FluentKey],
    all_fluent_keys: dict[str, FluentKey],
    depend_keys: set[str] | None = None,
) -> None:
    expression = placeable.expression

    if isinstance(expression, ast.VariableReference):
        _extract_kwargs_from_variable_reference(variable_reference=expression, kwargs=kwargs)

    elif isinstance(expression, ast.SelectExpression):
        _extract_kwargs_from_selector_expression(
            key=key,
            selector_expression=expression,
            kwargs=kwargs,
            terms=terms,
            all_fluent_keys=all_fluent_keys,
            depend_keys=depend_keys,
        )

    elif isinstance(expression, ast.MessageReference):
        # Add `ast.MessageReference.id.name` to depends_on_keys
        # to avoid key to be removed
        key.depends_on_keys.add(expression.id.name)
        if depend_keys is not None:
            depend_keys.add(expression.id.name)

        # Extract kwargs
        _extract_kwargs_from_message_reference(
            key=key,
            message_reference=expression,
            kwargs=kwargs,
            terms=terms,
            all_fluent_keys=all_fluent_keys,
            depend_keys=depend_keys,
        )

    elif isinstance(expression, ast.TermReference):
        _extract_kwargs_from_term_reference(
            key=key,
            term_expression=expression,
            kwargs=kwargs,
            terms=terms,
            all_fluent_keys=all_fluent_keys,
        )


def extract_kwargs(
    *,
    key: FluentKey,
    terms: dict[str, FluentKey] | None = None,
    all_fluent_keys: dict[str, FluentKey] | None = None,
    depend_keys: set[str] | None = None,
) -> set[str]:
    kwargs: set[str] = set()
    terms = terms or {}
    all_fluent_keys = all_fluent_keys or {}
    depend_keys = depend_keys if depend_keys is not None else set()

    if not isinstance(key.translation, (ast.Message, ast.Term)):
        return kwargs

    if not key.translation.value:
        return kwargs

    for element in key.translation.value.elements:
        if isinstance(element, ast.Placeable):
            _extract_kwargs_from_placeable(
                key=key,
                placeable=element,
                kwargs=kwargs,
                terms=terms,
                all_fluent_keys=all_fluent_keys,
                depend_keys=depend_keys,
            )

    return kwargs



from __future__ import annotations

from dataclasses import dataclass, field

from fluent.syntax import ast
from fluent.syntax.visitor import Visitor

from ftl_extract.stub.utils import remove_duplicates


class FTLStubCantFindMessageReferenceError(Exception): ...


class FTLStubCantFindTermError(Exception): ...


@dataclass
class Message:
    fluent_message: ast.Message
    kwargs: list[str] = field(default_factory=list)


@dataclass
class Term:
    fluent_term: ast.Term
    kwargs: list[str] = field(default_factory=list)


class FluentVisitor(Visitor):
    def __init__(self) -> None:
        self.messages: dict[str, Message] = {}  # { key: Message }
        self.terms: dict[str, Term] = {}  # { key: Term }

    @staticmethod
    def _extract_kwargs_from_variable_reference(
        *,
        obj: Message | Term,
        variable_reference: ast.VariableReference,
    ) -> None:
        obj.kwargs.append(variable_reference.id.name)

    def _extract_kwargs_from_selector_expression(
        self,
        *,
        obj: Message | Term,
        selector_expression: ast.SelectExpression,
    ) -> None:
        if isinstance(selector_expression.selector, ast.VariableReference):
            self._extract_kwargs_from_variable_reference(
                obj=obj,
                variable_reference=selector_expression.selector,
            )

        for variant in selector_expression.variants:
            for element in variant.value.elements:
                if isinstance(element, ast.Placeable):
                    self._extract_kwargs_from_placeable(
                        obj=obj,
                        placeable=element,
                    )

    def _extract_kwargs_from_message_reference(
        self,
        *,
        obj: Message | Term,
        message_reference: ast.MessageReference,
    ) -> None:
        reference_message = self.messages.get(message_reference.id.name, None)

        if not reference_message:
            raise FTLStubCantFindMessageReferenceError

        obj.kwargs.extend(self._extract_kwargs_from_message(message=reference_message))

    def _extract_kwargs_from_term_reference(
        self,
        *,
        obj: Message | Term,
        term_reference: ast.TermReference,
    ) -> None:
        term = self.terms.get(term_reference.id.name, None)

        if not term:
            raise FTLStubCantFindTermError

        obj.kwargs.extend(self._extract_kwargs_from_term(term=term))

    def _extract_kwargs_from_placeable(
        self,
        *,
        obj: Message | Term,
        placeable: ast.Placeable,
    ) -> None:
        expression = placeable.expression

        if isinstance(expression, ast.VariableReference):
            self._extract_kwargs_from_variable_reference(
                obj=obj,
                variable_reference=expression,
            )

        elif isinstance(expression, ast.SelectExpression):
            self._extract_kwargs_from_selector_expression(
                obj=obj,
                selector_expression=expression,
            )

        elif isinstance(expression, ast.MessageReference):
            self._extract_kwargs_from_message_reference(
                obj=obj,
                message_reference=expression,
            )

        elif isinstance(expression, ast.TermReference):
            self._extract_kwargs_from_term_reference(
                obj=obj,
                term_reference=expression,
            )

    def _extract_kwargs_from_message(self, message: Message) -> list[str]:
        if not message.fluent_message.value:
            return []

        for element in message.fluent_message.value.elements:
            if isinstance(element, ast.Placeable):
                self._extract_kwargs_from_placeable(
                    obj=message,
                    placeable=element,
                )

        message.kwargs = remove_duplicates(message.kwargs)
        return message.kwargs

    def _extract_kwargs_from_term(self, term: Term) -> list[str]:
        if not term.fluent_term.value:
            return []

        for element in term.fluent_term.value.elements:
            if isinstance(element, ast.Placeable):
                self._extract_kwargs_from_placeable(
                    obj=term,
                    placeable=element,
                )

        term.kwargs = remove_duplicates(term.kwargs)
        return term.kwargs

    def visit_Message(self, fluent_message: ast.Message) -> None:  # noqa: N802
        message = self.messages.setdefault(
            fluent_message.id.name,
            Message(fluent_message=fluent_message),
        )

        if not fluent_message.value:
            return self.generic_visit(fluent_message)

        self._extract_kwargs_from_message(message)

        return self.generic_visit(fluent_message)

    def visit_Term(self, fluent_term: ast.Term) -> None:  # noqa: N802
        term = self.terms.setdefault(fluent_term.id.name, Term(fluent_term=fluent_term))

        if not fluent_term.value:
            return self.generic_visit(fluent_term)

        self._extract_kwargs_from_term(term)

        return self.generic_visit(fluent_term)



from __future__ import annotations

from os import environ
from pprint import pformat
from typing import TYPE_CHECKING

from ftl_extract.const import FTL_DEBUG_VAR_NAME
from ftl_extract.utils import to_json_no_span

if TYPE_CHECKING:
    from pathlib import Path

    from fluent.syntax import ast


class FTLExtractorError(Exception):
    pass


class FTLExtractorDifferentPathsError(FTLExtractorError):
    def __init__(self, key: str, current_path: Path, new_path: Path) -> None:
        self.current_path = current_path
        self.new_path = new_path
        super().__init__(
            f"Key {key!r} already exists with different path: "
            f"{self.current_path} != {self.new_path}",
        )


class FTLExtractorDifferentTranslationError(FTLExtractorError):
    def __init__(
        self,
        key: str,
        current_translation: ast.Message,
        new_translation: ast.Message,
    ) -> None:
        self.current_translation = current_translation
        self.new_translation = new_translation

        if bool(environ.get(FTL_DEBUG_VAR_NAME, "")) is True:
            super().__init__(
                f"Translation {key!r} already exists with different elements:\n"
                f"current_translation: "
                f"{pformat(self.current_translation.to_json(fn=to_json_no_span))}\n!= "
                f"new_translation: {pformat(self.new_translation.to_json(fn=to_json_no_span))}",
            )
        else:
            super().__init__(
                f"Translation {key!r} already exists with different elements: "
                f"{self.current_translation} != {self.new_translation}",
            )


class FTLExtractorCantFindReferenceError(FTLExtractorError):
    def __init__(self, key: str, key_path: Path, reference_key: str) -> None:
        self.key = key
        self.key_path = key_path
        self.reference_key = reference_key
        super().__init__(f"Can't find reference {reference_key!r} for key {key!r} at {key_path}")


class FTLExtractorCantFindTermError(FTLExtractorError):
    def __init__(self, key: str, locale: str, key_path: Path, term_key: str) -> None:
        self.key = key
        self.locale = locale
        self.key_path = key_path
        self.term_key = term_key
        super().__init__(
            f"Can't find term {term_key!r} for key {key!r} with locale {locale!r} at: {key_path}",
        )



from __future__ import annotations

import ast
import json
from typing import TYPE_CHECKING, cast

import click
from fluent.syntax import FluentParser

from ftl_extract.stub.tree import METADATA_DICT_KEY, Metadata, generate_tree
from ftl_extract.stub.utils import to_camel_case
from ftl_extract.stub.visitor import FluentVisitor, Message

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path
    from typing import Any


class NoBodyError(Exception): ...


class RootKeyIsMissingError(Exception): ...


def read_ftl_messages(visitor: FluentVisitor, path: Path) -> dict[str, Message]:
    resource = FluentParser().parse(path.read_text(encoding="utf-8"))
    if resource.body is None:
        msg = "no body"
        raise NoBodyError(msg)

    visitor.visit(resource)

    return visitor.messages


def locate_ftl_files(path: Path) -> Generator[Path, Any, Any]:
    if path.is_dir():
        yield from path.rglob("*.ftl")
    else:
        yield from (path,)


def build_base_ast() -> ast.Module:
    return ast.Module(
        body=[
            ast.ImportFrom(module="collections.abc", names=[ast.alias(name="Generator")], level=0),
            ast.ImportFrom(module="contextlib", names=[ast.alias(name="contextmanager")], level=0),
            ast.ImportFrom(
                module="typing",
                names=[
                    ast.alias(name="Any"),
                    ast.alias(name="Literal"),
                    ast.alias(name="overload"),
                ],
                level=0,
            ),
            ast.ImportFrom(module="aiogram_i18n", names=[ast.alias(name="LazyProxy")], level=0),
            ast.ClassDef(
                name="I18nContext",
                bases=[ast.Name(id="I18nStub")],
                keywords=[],
                body=[
                    ast.FunctionDef(
                        name="get",
                        args=ast.arguments(
                            posonlyargs=[
                                ast.arg(arg="self"),
                                ast.arg(arg="key", annotation=ast.Name(id="str")),
                            ],
                            args=[],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=ast.arg(arg="kwargs", annotation=ast.Name(id="Any")),
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[],
                        returns=ast.Name(id="str"),
                    ),
                    ast.AsyncFunctionDef(
                        name="set_locale",
                        args=ast.arguments(
                            posonlyargs=[],
                            args=[
                                ast.arg(arg="self"),
                                ast.arg(arg="locale", annotation=ast.Name(id="str")),
                            ],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=ast.arg(arg="kwargs", annotation=ast.Name(id="Any")),
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[],
                        returns=ast.Constant(value=None),
                    ),
                    ast.FunctionDef(
                        name="use_locale",
                        args=ast.arguments(
                            posonlyargs=[],
                            args=[
                                ast.arg(arg="self"),
                                ast.arg(arg="locale", annotation=ast.Name(id="str")),
                            ],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=None,
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[ast.Name(id="contextmanager")],
                        returns=ast.Subscript(
                            value=ast.Name(id="Generator"),
                            slice=ast.Tuple(
                                elts=[
                                    ast.Name(id="I18nContext"),
                                    ast.Constant(value=None),
                                    ast.Constant(value=None),
                                ],
                            ),
                        ),
                    ),
                    ast.FunctionDef(
                        name="use_context",
                        args=ast.arguments(
                            posonlyargs=[],
                            args=[ast.arg(arg="self")],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=ast.arg(arg="kwargs", annotation=ast.Name(id="Any")),
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[ast.Name(id="contextmanager")],
                        returns=ast.Subscript(
                            value=ast.Name(id="Generator"),
                            slice=ast.Tuple(
                                elts=[
                                    ast.Name(id="I18nContext"),
                                    ast.Constant(value=None),
                                    ast.Constant(value=None),
                                ],
                            ),
                        ),
                    ),
                    ast.FunctionDef(
                        name="set_context",
                        args=ast.arguments(
                            posonlyargs=[],
                            args=[ast.arg(arg="self")],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=ast.arg(arg="kwargs", annotation=ast.Name(id="Any")),
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[],
                        returns=ast.Constant(value=None),
                    ),
                ],
                decorator_list=[],
            ),
            ast.ClassDef(
                name="LazyFactory",
                bases=[ast.Name(id="I18nStub")],
                keywords=[],
                body=[
                    ast.AnnAssign(
                        target=ast.Name(id="key_separator"),
                        annotation=ast.Name(id="str"),
                        simple=1,
                    ),
                    ast.FunctionDef(
                        name="set_separator",
                        args=ast.arguments(
                            posonlyargs=[],
                            args=[
                                ast.arg(arg="self"),
                                ast.arg(arg="key_separator", annotation=ast.Name(id="str")),
                            ],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=None,
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[],
                        returns=ast.Constant(value=None),
                    ),
                    ast.FunctionDef(
                        name="__call__",
                        args=ast.arguments(
                            posonlyargs=[
                                ast.arg(arg="self"),
                                ast.arg(arg="key", annotation=ast.Name(id="str")),
                            ],
                            args=[],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=ast.arg(
                                arg="kwargs",
                                annotation=ast.Subscript(
                                    value=ast.Name(id="dict"),
                                    slice=ast.Tuple(
                                        elts=[
                                            ast.Name(id="str"),
                                            ast.Name(id="Any"),
                                        ],
                                    ),
                                ),
                            ),
                            defaults=[],
                        ),
                        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
                        decorator_list=[],
                        returns=ast.Name(id="LazyProxy"),
                    ),
                ],
                decorator_list=[],
            ),
            ast.AnnAssign(target=ast.Name(id="L"), annotation=ast.Name(id="LazyFactory"), simple=1),
        ],
        type_ignores=[],
    )


def create_static_method(name: str, metadata: Metadata) -> ast.FunctionDef:
    args = ast.arguments(
        posonlyargs=[],
        args=[],
        vararg=None,
        kwonlyargs=[
            ast.arg(arg=arg_name, annotation=ast.Name(id="Any", ctx=ast.Load()))
            for arg_name in metadata["args"]
        ],
        kw_defaults=[None for _ in metadata["args"]],
        kwarg=ast.arg(arg="kwargs", annotation=ast.Name(id="Any", ctx=ast.Load())),
        defaults=[],
    )
    return ast.FunctionDef(
        name=name,
        args=args,
        decorator_list=[ast.Name(id="staticmethod", ctx=ast.Load())],
        returns=ast.Subscript(
            value=ast.Name(id="Literal", ctx=ast.Load()),
            slice=ast.Constant(value=metadata["translation"].split("\n", maxsplit=1)[0]),
            ctx=ast.Load(),
        ),
        body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
        type_comment=None,
    )


def process_tree(
    name: str,
    tree: dict[str, Any],
    parent_body: list[ast.stmt],
) -> None:
    if METADATA_DICT_KEY in tree:
        static_method = create_static_method(name, cast(Metadata, tree.pop(METADATA_DICT_KEY)))
        parent_body.append(static_method)

        if tree:
            static_method.decorator_list.append(ast.Name(id="overload", ctx=ast.Load()))

    if tree:
        class_def = ast.ClassDef(
            name=f"__{to_camel_case(name)}",
            bases=[],
            keywords=[],
            body=[],
            decorator_list=[],
        )

        for key, value in tree.items():
            process_tree(key, value, class_def.body)

        parent_body.append(cast(ast.stmt, class_def))

        parent_body.append(
            cast(
                ast.stmt,
                ast.Assign(
                    targets=[ast.Name(id=name, ctx=ast.Store())],
                    value=ast.Call(
                        func=ast.Name(id=f"__{to_camel_case(name)}", ctx=ast.Load()),
                        args=[],
                        keywords=[],
                    ),
                ),
            ),
        )


def generate_ast(module: ast.Module, tree: dict[str, dict[str, Any]]) -> None:
    if "i18n_stub" not in tree:
        msg = "i18n_stub key is missing in the tree"
        raise RootKeyIsMissingError(msg)

    top_class = ast.ClassDef(
        name="I18nStub",
        bases=[],
        keywords=[],
        body=[],
        decorator_list=[],
    )
    for key, value in tree["i18n_stub"].items():
        if key == METADATA_DICT_KEY:
            continue

        process_tree(key, value, top_class.body)

    module.body.append(top_class)


def generate_stubs(ftl_path: Path, output_path: Path, export_tree: bool = False) -> None:
    if output_path.is_dir():
        output_path /= "stub.pyi"

    if output_path.suffix != ".pyi":
        msg = f"Output file `{output_path.name}` must have `.pyi` extension"
        raise ValueError(msg)

    ftl_files = locate_ftl_files(ftl_path)
    visitor = FluentVisitor()

    for ftl_file in ftl_files:
        read_ftl_messages(visitor, ftl_file)

    tree = generate_tree(visitor.messages)
    tree: dict[str, dict[str, Any]] = {"i18n_stub": {**tree}}
    if export_tree:
        (output_path.parent / "stub.json").write_text(
            json.dumps(tree, indent=2, ensure_ascii=False),
            encoding="utf-8",
        )
        click.echo(f"Tree structure exported to {output_path.parent / 'stub.json'}")

    module = build_base_ast()
    generate_ast(module, tree)
    ast.fix_missing_locations(module)

    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text("# mypy: ignore-errors\n" + ast.unparse(module), encoding="utf-8")

    click.echo(f"Stub file generated at {output_path}")



from __future__ import annotations

from os import linesep
from typing import TYPE_CHECKING

from click import echo
from fluent.syntax import FluentSerializer
from fluent.syntax import ast as fl_ast

from ftl_extract import extract_fluent_keys
from ftl_extract.code_extractor import sort_fluent_keys_by_path
from ftl_extract.const import (
    COMMENT_KEYS_MODE,
    DEFAULT_EXCLUDE_DIRS,
    DEFAULT_FTL_FILE,
    DEFAULT_I18N_KEYS,
    DEFAULT_IGNORE_ATTRIBUTES,
    DEFAULT_IGNORE_KWARGS,
    LINE_ENDINGS,
)
from ftl_extract.ftl_importer import import_ftl_from_dir
from ftl_extract.process.commentator import comment_ftl_key
from ftl_extract.process.kwargs_extractor import extract_kwargs
from ftl_extract.process.serializer import generate_ftl
from ftl_extract.utils import ExtractionStatistics, prepare_exclude_dirs

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from pathlib import Path

    from ftl_extract.matcher import FluentKey


def extract(
    *,
    code_path: Path,
    output_path: Path,
    language: Sequence[str],
    i18n_keys: Iterable[str] = DEFAULT_I18N_KEYS,
    i18n_keys_append: Iterable[str] = (),
    i18n_keys_prefix: Iterable[str] = (),
    exclude_dirs: tuple[str, ...] = DEFAULT_EXCLUDE_DIRS,
    exclude_dirs_append: tuple[str, ...] = (),
    ignore_attributes: Iterable[str] = DEFAULT_IGNORE_ATTRIBUTES,
    append_ignore_attributes: Iterable[str] = (),
    ignore_kwargs: Iterable[str] = DEFAULT_IGNORE_KWARGS,
    comment_junks: bool = True,
    default_ftl_file: Path = DEFAULT_FTL_FILE,
    comment_keys_mode: str = COMMENT_KEYS_MODE[0],
    line_endings: str = LINE_ENDINGS[0],
    serializer: FluentSerializer | None = None,
    dry_run: bool = False,
    verbose: bool = False,
) -> ExtractionStatistics:
    statistics = ExtractionStatistics()
    statistics.ftl_files_count = dict.fromkeys(language, 0)
    statistics.ftl_stored_keys_count = dict.fromkeys(language, 0)
    statistics.ftl_keys_updated = dict.fromkeys(language, 0)
    statistics.ftl_keys_added = dict.fromkeys(language, 0)
    statistics.ftl_keys_commented = dict.fromkeys(language, 0)

    exclude_dirs = prepare_exclude_dirs(
        exclude_dirs=exclude_dirs,
        exclude_dirs_append=exclude_dirs_append,
    )

    if i18n_keys_append:
        i18n_keys = (*i18n_keys, *i18n_keys_append)

    if append_ignore_attributes:
        ignore_attributes = (*ignore_attributes, *append_ignore_attributes)

    if serializer is None:
        serializer = FluentSerializer(with_junk=True)

    if line_endings == "default":
        line_endings = linesep
    elif line_endings == "lf":
        line_endings = "\n"
    elif line_endings == "cr":
        line_endings = "\r"
    elif line_endings == "crlf":
        line_endings = "\r\n"
    else:
        msg = f"Invalid line endings: {line_endings!r}"
        raise ValueError(msg)

    # Extract fluent keys from code
    in_code_fluent_keys = extract_fluent_keys(
        path=code_path,
        i18n_keys=i18n_keys,
        i18n_keys_prefix=i18n_keys_prefix,
        exclude_dirs=exclude_dirs,
        ignore_attributes=ignore_attributes,
        ignore_kwargs=ignore_kwargs,
        default_ftl_file=default_ftl_file,
        statistics=statistics,
    )
    statistics.ftl_in_code_keys_count = len(in_code_fluent_keys)

    for lang in language:
        # Import fluent keys and terms from existing FTL files
        stored_fluent_keys, stored_terms, leave_as_is = import_ftl_from_dir(
            path=output_path,
            locale=lang,
            statistics=statistics,
        )
        for fluent_key in stored_fluent_keys.values():
            fluent_key.path = fluent_key.path.relative_to(output_path / lang)

        for term in stored_terms.values():
            term.path = term.path.relative_to(output_path / lang)

        keys_to_comment: dict[str, FluentKey] = {}
        keys_to_add: dict[str, FluentKey] = {}

        # Find keys should be commented
        # Keys, that are not in code or not in their `path_` file
        # First step: find keys that have different paths
        for key, fluent_key in in_code_fluent_keys.items():
            if key in stored_fluent_keys and fluent_key.path != stored_fluent_keys[key].path:
                keys_to_comment[key] = stored_fluent_keys.pop(key)
                statistics.ftl_keys_commented[lang] += 1

                keys_to_add[key] = fluent_key
                statistics.ftl_keys_updated[lang] += 1

            elif key not in stored_fluent_keys:
                keys_to_add[key] = fluent_key
                statistics.ftl_keys_added[lang] += 1

            else:
                stored_fluent_keys[key].code_path = fluent_key.code_path

        # Second step: find keys that have different kwargs

        # Keys that are not in code but stored keys are depends on them
        depend_keys: set[str] = set()

        for key, fluent_key in in_code_fluent_keys.items():
            if key not in stored_fluent_keys:
                continue

            fluent_key_placeable_set = extract_kwargs(
                key=fluent_key,
                terms=stored_terms,
                all_fluent_keys=in_code_fluent_keys.copy(),
                depend_keys=depend_keys,
            )
            stored_fluent_key_placeable_set = extract_kwargs(
                key=stored_fluent_keys[key],
                terms=stored_terms,
                all_fluent_keys=stored_fluent_keys.copy(),
                depend_keys=depend_keys,
            )

            if fluent_key_placeable_set != stored_fluent_key_placeable_set:
                keys_to_comment[key] = stored_fluent_keys.pop(key)
                statistics.ftl_keys_commented[lang] += 1

                keys_to_add[key] = fluent_key
                statistics.ftl_keys_updated[lang] += 1

        # Third step: find keys that are not in code
        for key in stored_fluent_keys.keys() - in_code_fluent_keys.keys():
            if key in depend_keys:
                continue

            keys_to_comment[key] = stored_fluent_keys.pop(key)
            statistics.ftl_keys_commented[lang] += 1

        if comment_keys_mode == "comment":
            for fluent_key in keys_to_comment.values():
                comment_ftl_key(key=fluent_key, serializer=serializer)

        elif comment_keys_mode == "warn":
            for fluent_key in keys_to_comment.values():
                keys_to_add.pop(fluent_key.key, None)
                echo(
                    f"Key `{fluent_key.key}` with such kwargs in "
                    f"`{output_path / lang / fluent_key.path}` is not in code.",
                )

        # Comment Junk elements if needed
        if comment_junks is True:
            for fluent_key in leave_as_is:
                if isinstance(fluent_key.translation, fl_ast.Junk):
                    comment_ftl_key(key=fluent_key, serializer=serializer)
                    statistics.ftl_keys_commented[lang] += 1

        sorted_fluent_keys = sort_fluent_keys_by_path(fluent_keys=stored_fluent_keys)

        for path, keys in sort_fluent_keys_by_path(fluent_keys=keys_to_add).items():
            sorted_fluent_keys.setdefault(path, []).extend(keys)

        for path, keys in sort_fluent_keys_by_path(fluent_keys=keys_to_comment).items():
            sorted_fluent_keys.setdefault(path, []).extend(keys)

        for path, keys in sort_fluent_keys_by_path(fluent_keys=stored_terms).items():
            sorted_fluent_keys.setdefault(path, []).extend(keys)

        leave_as_is_with_path: dict[Path, list[FluentKey]] = {}

        for fluent_key in leave_as_is:
            leave_as_is_with_path.setdefault(
                fluent_key.path.relative_to(output_path / lang),
                [],
            ).append(fluent_key)

        for path, keys in sorted_fluent_keys.items():
            ftl, _ = generate_ftl(
                fluent_keys=keys,
                serializer=serializer,
                leave_as_is=leave_as_is_with_path.get(path, []),
            )
            if dry_run:
                if verbose:
                    echo(
                        f"[DRY-RUN] File {output_path / lang / path} has been saved. {len(keys)} "
                        f"keys found.",
                    )
            else:
                _write(path=output_path / lang / path, ftl=ftl, line_endings=line_endings)
                if verbose:
                    echo(
                        f"File {output_path / lang / path} has been saved. {len(keys)} keys found."
                    )

            statistics.ftl_stored_keys_count[lang] += len(
                [key for key in keys if isinstance(key.translation, fl_ast.Message)],
            )

    return statistics


def _write(*, path: Path, ftl: str, line_endings: str) -> None:
    """Write FTL to file."""
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(ftl, encoding="utf-8", newline=line_endings)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Final

I18N_LITERAL: Final[str] = "i18n"
GET_LITERAL: Final[str] = "get"
PATH_LITERAL: Final[str] = "_path"
DEFAULT_I18N_KEYS: Final[tuple[str, ...]] = (I18N_LITERAL, "L", "LazyProxy", "LazyFilter")
DEFAULT_IGNORE_ATTRIBUTES: Final[tuple[str, ...]] = (
    "set_locale",
    "use_locale",
    "use_context",
    "set_context",
)
DEFAULT_IGNORE_KWARGS: Final[tuple[str, ...]] = ()
DEFAULT_FTL_FILE: Final[Path] = Path("_default.ftl")
FTL_DEBUG_VAR_NAME: Final[str] = "FTL_DEBUG"
COMMENT_KEYS_MODE: Final[tuple[str, ...]] = ("comment", "warn")
DEFAULT_EXCLUDE_DIRS: Final[tuple[str, ...]] = (
    ".venv",
    "venv",
    ".git",
    "__pycache__",
    ".pytest_cache",
)
LINE_ENDINGS: Final[tuple[str, ...]] = ("default", "lf", "cr", "crlf")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any


def to_camel_case(string: str) -> str:
    return "".join(x.capitalize() for x in string.lower().split("_"))


def remove_duplicates(sequence: Sequence[Any]) -> list[Any]:
    exists_set = set()
    return [x for x in sequence if not (x in exists_set or exists_set.add(x))]  # type: ignore[func-returns-value]



from .stub import I18nContext

i18n = I18nContext()

# Text
i18n.text()

# Text with arguments
i18n.text.kwargs(kwarg1="value1", kwarg2="value2")

# Text with arguments as term
i18n.text.args.term()

# Text with arguments as term with arguments in terms
i18n.text.args.term.args(kwarg1="value1", kwarg2="value2")

# Text with message_reference
i18n.text.message_reference()

# Text with message_reference with arguments
i18n.text.message_reference.args(kwarg1="value1", kwarg2="value2")

# Text with selector
i18n.text.selector(selector=1)

# Text with selector with selectors
i18n.text.selector.selectors(selector=1)

# Text with selector with arguments
i18n.text.selector.kwargs(selector=1, kwarg1="value1", kwarg2="value2")

# Text with selector with selectors and arguments
i18n.text.selector.reference.selector.kwargs.terms(
    selector=1,
    kwarg1="value1",
    kwarg2="value2",
)



from __future__ import annotations

import subprocess
import sys
from pathlib import Path
from time import perf_counter_ns
from typing import Literal

import click

from ftl_extract.const import (
    COMMENT_KEYS_MODE,
    DEFAULT_EXCLUDE_DIRS,
    DEFAULT_FTL_FILE,
    DEFAULT_I18N_KEYS,
    DEFAULT_IGNORE_ATTRIBUTES,
    DEFAULT_IGNORE_KWARGS,
    LINE_ENDINGS,
)
from ftl_extract.ftl_extractor import extract
from ftl_extract.stub.generator import generate_stubs


@click.group("ftl")
@click.version_option()
def ftl() -> None: ...


@ftl.command("extract")
@click.argument("code_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
@click.option(
    "--language",
    "-l",
    multiple=True,
    default=("en",),
    show_default=True,
    help="Language of translation.",
)
@click.option(
    "--i18n-keys",
    "-k",
    default=DEFAULT_I18N_KEYS,
    multiple=True,
    show_default=True,
    help="Names of function that is used to get translation.",
)
@click.option(
    "--i18n-keys-append",
    "-K",
    default=(),
    multiple=True,
    help="Append names of function that is used to get translation.",
)
@click.option(
    "--i18n-keys-prefix",
    "-p",
    default=(),
    multiple=True,
    help="Prefix names of function that is used to get translation. `self.i18n.*()`",
)
@click.option(
    "--exclude-dirs",
    "-e",
    multiple=True,
    default=DEFAULT_EXCLUDE_DIRS,
    show_default=True,
    help="Exclude directories.",
)
@click.option(
    "--exclude-dirs-append",
    "-E",
    default=(),
    multiple=True,
    help="Append directories to exclude.",
)
@click.option(
    "--ignore-attributes",
    "-i",
    default=DEFAULT_IGNORE_ATTRIBUTES,
    multiple=True,
    show_default=True,
    help="Ignore attributes, like `i18n.set_locale`.",
)
@click.option(
    "--append-ignore-attributes",
    "-I",
    multiple=True,
    help="Append attributes to ignore.",
)
@click.option(
    "--ignore-kwargs",
    default=DEFAULT_IGNORE_KWARGS,
    multiple=True,
    show_default=True,
    help="Ignore kwargs, like `when` from `aiogram_dialog.I18nFormat(..., when=...)`.",
)
@click.option(
    "--comment-junks",
    is_flag=True,
    default=False,
    show_default=True,
    help="Comments Junk elements.",
)
@click.option(
    "--default-ftl-file",
    default=DEFAULT_FTL_FILE,
    show_default=True,
    type=click.Path(path_type=Path),
)
@click.option(
    "--comment-keys-mode",
    default=COMMENT_KEYS_MODE[0],
    show_default=True,
    help="Comment keys mode.",
    type=click.Choice(COMMENT_KEYS_MODE, case_sensitive=False),
)
@click.option(
    "--line-endings",
    default=LINE_ENDINGS[0],
    show_default=True,
    type=click.Choice(LINE_ENDINGS, case_sensitive=False),
    help="Line endings for generated FTL files.",
)
@click.option(
    "--fast",
    is_flag=True,
    default=False,
    show_default=True,
    help="Run the fast Rust implementation.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    show_default=True,
    help="Do not write to output files.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    show_default=True,
    help="Verbose output.",
)
def cli_extract(
    code_path: Path,
    output_path: Path,
    language: tuple[str, ...],
    i18n_keys: tuple[str, ...],
    i18n_keys_append: tuple[str, ...],
    i18n_keys_prefix: tuple[str, ...],
    exclude_dirs: tuple[str, ...],
    exclude_dirs_append: tuple[str, ...],
    ignore_attributes: tuple[str, ...],
    append_ignore_attributes: tuple[str, ...],
    ignore_kwargs: tuple[str, ...],
    comment_junks: bool,
    default_ftl_file: Path,
    comment_keys_mode: Literal["comment", "warn"],
    line_endings: Literal["default", "lf", "cr", "crlf"],
    fast: bool,
    dry_run: bool,
    verbose: bool,
) -> None:
    if fast:
        cmd = ["fast-ftl"]

        if verbose:
            cmd.append("--verbose")

        cmd.extend(["extract", str(code_path), str(output_path)])

        # Add multi-value options to the command
        multi_value_options = {
            "--language": language,
            "--i18n-keys": i18n_keys,
            "--i18n-keys-append": i18n_keys_append,
            "--i18n-keys-prefix": i18n_keys_prefix,
            "--exclude-dirs": exclude_dirs,
            "--exclude-dirs-append": exclude_dirs_append,
            "--ignore-attributes": ignore_attributes,
            "--append-ignore-attributes": append_ignore_attributes,
            "--ignore-kwargs": ignore_kwargs,
        }
        for option, values in multi_value_options.items():
            for value in values:
                cmd.extend([option, value])

        # Add single-value options to the command
        cmd.extend(
            [
                "--default-ftl-file",
                str(default_ftl_file),
                "--comment-keys-mode",
                comment_keys_mode,
                "--line-endings",
                line_endings,
            ]
        )

        # Boolean flags
        for flag, condition in {"--comment-junks": comment_junks, "--dry-run": dry_run}.items():
            if condition:
                cmd.append(flag)

        try:
            result = subprocess.run(cmd, check=True)  # noqa: S603
        except FileNotFoundError:
            click.secho("Error: 'fast-ftl-extract' executable not found.", fg="red")
            click.secho(
                "Please ensure the Rust implementation is built and in your PATH.", fg="red"
            )
            sys.exit(1)
        except subprocess.CalledProcessError as e:
            # The Rust process failed. Its error message was already printed.
            # We exit with the same error code.
            sys.exit(e.returncode)
        else:
            sys.exit(result.returncode)

    click.echo(f"Extracting from {code_path}")
    start_time = perf_counter_ns()

    statistics = extract(
        code_path=code_path,
        output_path=output_path,
        language=language,
        i18n_keys=i18n_keys,
        i18n_keys_append=i18n_keys_append,
        i18n_keys_prefix=i18n_keys_prefix,
        exclude_dirs=exclude_dirs,
        exclude_dirs_append=exclude_dirs_append,
        ignore_attributes=ignore_attributes,
        append_ignore_attributes=append_ignore_attributes,
        ignore_kwargs=ignore_kwargs,
        comment_junks=comment_junks,
        default_ftl_file=default_ftl_file,
        comment_keys_mode=comment_keys_mode,
        line_endings=line_endings,
        dry_run=dry_run,
        verbose=verbose,
    )

    if verbose:
        click.echo("Extraction statistics:")
        click.echo(f"  - Py files count: {statistics.py_files_count}")
        click.echo(f"  - FTL files count: {statistics.ftl_files_count}")
        click.echo(f"  - FTL keys in code: {statistics.ftl_in_code_keys_count}")
        click.echo(f"  - FTL keys stored: {statistics.ftl_stored_keys_count}")
        click.echo(f"  - FTL keys updated: {statistics.ftl_keys_updated}")
        click.echo(f"  - FTL keys added: {statistics.ftl_keys_added}")
        click.echo(f"  - FTL keys commented: {statistics.ftl_keys_commented}")

    click.echo(f"[Python] Done in {(perf_counter_ns() - start_time) * 1e-9:.3f}s.")


@ftl.command("stub")
@click.argument("locale_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
@click.option(
    "--export-tree",
    is_flag=True,
    default=False,
    show_default=True,
    help="Export tree structure of FTL messages.",
)
def cli_stub(locale_path: Path, output_path: Path, export_tree: bool) -> None:
    click.echo(f"Generating stubs from {locale_path}")
    start_time = perf_counter_ns()

    generate_stubs(locale_path, output_path, export_tree)

    click.echo(f"[Python] Done in {(perf_counter_ns() - start_time) * 1e-9:.3f}s.")



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any


def to_json_no_span(value: dict[str, Any]) -> Any:
    value.pop("span", None)
    return value


@dataclass
class ExtractionStatistics:
    py_files_count: int = field(default=0)
    ftl_files_count: dict[str, int] = field(default_factory=dict)  # dict[lang, count]
    ftl_in_code_keys_count: int = field(default=0)
    ftl_stored_keys_count: dict[str, int] = field(default_factory=dict)  # dict[lang, count]
    ftl_keys_updated: dict[str, int] = field(default_factory=dict)  # dict[lang, count]
    ftl_keys_added: dict[str, int] = field(default_factory=dict)  # dict[lang, count]
    ftl_keys_commented: dict[str, int] = field(default_factory=dict)  # dict[lang, count]


def prepare_exclude_dirs(
    *,
    exclude_dirs: tuple[str, ...],
    exclude_dirs_append: tuple[str, ...],
) -> frozenset[Path]:
    return frozenset(
        [Path(exclude_dir) for exclude_dir in exclude_dirs + exclude_dirs_append],
    )



from os import getenv

from setuptools import setup
from setuptools_rust import RustBin, Strip

rust_extensions = []

# Read environment variable `BUILD_RUST_IMPL`, if set to "1" build the Rust implementation
if getenv("BUILD_RUST_IMPL", "0") == "1":
    rust_extensions.append(
        RustBin(
            "fast-ftl",
            path="src/cli/Cargo.toml",
            strip=Strip.All,
            debug=False,
        )
    )

setup(
    package_dir={"ftl_extract": "src/ftl_extract"},
    rust_extensions=rust_extensions,
)
